package example;

import java.text.ChoiceFormat;
import java.text.Collator;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.text.Format;
import java.text.NumberFormat;
import java.text.BreakIterator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

public class Main {
	/*
	 * Wprowadzenie.
	 * 
	 * Podczas tworzenia wielojzycznego interfejsu aplikacji wykorzystuje si kilka
	 * klas z pakietu java.util. Ponadto tworzy si pliki waciwoci bd pliki
	 * klas, osobno dla rnych lokalizacji. Obowizuje przy tym konwencja
	 * nazewnicza. Plik domylny ma nazw bez adnych przyrostkw, pliki specyficzne
	 * dla danej lokalizacj maj doczane do nazwy symbole jzyka i dialektu. 
	 * 
	 * W niniejszym projekcie pliki waciwoci to: 
	 *    MyBudle.properties, MyBundle_en_EN.properties, MyBundle_pl_PL.properties 
	 * Umieszczono je w folderze resources (znajdujcym si w ciece klas)
	 * 
	 * Pliki rdowe klas to: 
	 *    MyBudle.java, MyBundle_en_EN.java, MyBundle_pl_PL.java
	 * Umieszczono je w pakiecie resources, tak e po skompilowaniu wygenerowane klasy
	 * trafi do odpowiadajcego pakietowi katalogu resources w ciece klas.
	 * 
	 * W wymienionych plikach dostarczane s zestawy: 
	 *   klucz (identyfikator) oraz 
	 *   warto (cig znakw, ktry moe zawiera nawiasy klamrowe - tj. szablony).
	 * 
	 * Aplikacja na podstawie kluczy, ktre s niezalene od lokalizacji,
	 * moe pozyska cigi znakw waciwe dla danej lokalizacji. 
	 * 
	 * Jeli w cigach znakw s szablony, to da si w nie "wstrzykn" zadane wartoci.
	 * Ponadto te wartoci mona sformatowa. Da si te wykorzysta te cigi w celu 
	 * pozyskania wartoci (poprzez parsowanie).
	 * Zastosowania te uzupenia jeszcze moliwo formatowania dat i wartoci numerycznych.
	  
	 * Materiay do przestudiowania:
	 *  https://docs.oracle.com/javase/7/docs/api/java/text/ChoiceFormat.html
	 *  https://docs.oracle.com/javase/7/docs/api/java/text/MessageFormat.html
	 *
	 */
	
	
	public static void main(String[] args) {
		String language = "pl";
		String country = "PL";

		// Klasa reprezentujca lokalizacj
		Locale l = new Locale(language, country);

		/*
		 * Pliki z waciwociami (xxx.properties) powinny znajdowa si w ciece
		 * klas. Cigi znakw w nich zawarte obsugiwane s za pomoc klasy
		 * PropertyResourceBundle. Klasa ta dziedziczy z ResourceBundle. Odpowiednie
		 * instancje dostarcza metoda ResourceBundle.getBundle (zwraca
		 * PropertyResourceBundle).
		 */

		// Przykad (cigi znakw adowane z plikw z waciwociami)
		ResourceBundle rb1 = ResourceBundle.getBundle("MyBundle", l);
		System.out.println(rb1.getString("greetings"));
		System.out.println(rb1.getString("inquire"));
		System.out.println(rb1.getString("farewell"));

		/*
		 * Pliki klas (xxx.class) powstae po skompilowaniu plikw rdowych 
		 * powinny znale si w ciece klas. 
		 * Same klasy powinny dziedziczy po ListResourceBundle.
		 * Odpowiednie instancje dostarczy metoda ResourceBundle.getBundle (zwraca
		 * obiekty typu ListResourceBundle)
		 */

		// Przykad pobrania cigw znakw po kluczu (adowanych ze skompilowanych klas)
		ResourceBundle rb2 = ResourceBundle.getBundle("resources.MyBundle", l);
		System.out.println(rb2.getString("greetings"));
		System.out.println(rb2.getString("inquire"));
		System.out.println(rb2.getString("farewell"));

		/*
		 * Definicje zawarte w xxx.class maj wyszy priorytet ni definicje w plikach
		 * xxx.properties. Gdy w ciece klas wystpi tak samo nazwane bundle obu
		 * rozdzajw, to pierwszestwo uzyskaj pliki xxx.class.
		 * W przypadku braku zasobw z definicjami wyrzucany jest wyjtek 
		 * MissingResourceException. Istnieje te "dziedziczenie" definicji
		 * 
		 * Dziedziczenie (https://www.baeldung.com/java-resourcebundle)
		 * 
		 * Pary klucz-warto zamieszczone w plikach bazowych (znajdujce si wyej w
		 * drzewie) widziane s przy prbie pozyskania zasobw dla bardziej specyficznej
		 * lokalizacji. Jeli dane s np. trzy rne pliki:
		 * 
		 * #resource.properties cancelButton = cancel
		 * 
		 * #resource_pl.properties continueButton = dalej
		 * 
		 * #resource_pl_PL.properties backButton = cofnij
		 * 
		 * to przy prbie pozyskania stringw dla Locale("pl", "PL") powinno da si
		 * uzyska wszystkie trzy cigi znakw. W efekcie moe te nastpi "nadpisanie"
		 * cigw znakw (gdy w potomnych plikach pojawi si takie same klucze jak
		 * w plikach nadrzdnych.
		 * 
		 * Co ciekawe, ListResourceBundles oraz PropertyResourceBundles nie nale do
		 * tej samej hierarchii.
		 */

		/* Przykad zastosowania szablonu.
		 * 
		 * Instancja MessageFormat pozwala wstrzykn wartoci i je sformatowa.
		 * Wymaga ona ustawieni lokalizacji.
		 */
		
		MessageFormat mf = new MessageFormat("");
		mf.setLocale(l);

		// Po zdefiniowani tablicy parametrw:
		Object[] arguments = { rb1.getString("planet"), 7, new Date() };
		// i wczytaniu szablonu z bundla
		mf.applyPattern(rb1.getString("template"));
		// mona te parametry zaaplikowa celem sformatowania cigu znakw.
		// Co wane, liczba argumentw nie moe by mniejsza ni liczba argumentw 
		// wymienianych w szablonie, za liczby w klamrowych nawiasach w szablonie
		// peni rol indeksw.
		String output = mf.format(arguments);
		// Wynik mona np. wypisa na ekranie.
		System.out.println(output);

		/* Przykad zastosowania szablonu uzupenianego warunkowo
		 * https://docs.oracle.com/javase/7/docs/api/java/text/ChoiceFormat.html
		 *
		 * Potrzebna do tego bdzie instancja ChoiceFormat odpowiednio sparametryzowana
		 * Klasa ChoiceFormat pozwala uzaleni sposb formatowania od przynalenoci
		 * danej wartoci do okrelonego przedziau (tj. zdefiniowa mapowanie: jeli
		 * dana warto naley do j-tego przedziau, to powinien zosta wybrany j-ty
		 * cig znakw). Dziki wsppracy z MessageFormat mona za jej pomoc obsuy
		 * odmian przez liczby.
		 * 
		 * Mapowanie definiuje si przez podanie uporzdkowanej rosnco listy wartoci
		 * typu doubles. Kady element tej listy definiuje przedzia potwarty do
		 * nastpnego elementu.
		 * 
		 * warto X pasuje do j wtedy i tylko wtedy, gdy limit[j] <= X < limit[j+1]
		 * 
		 * W przypadku, gdy warto wychodzi poza obszar zdefiniowany przez te
		 * przedziay, to wybrany zostanie pierwszy lub ostatni indeks, zalenie od
		 * tego, czy (X) jest zbyt mae lub zbyt due.
		 */

		// Najpierw zdefiniowane zostan przedziay oraz odpowiadajce przedziaom
		// teksty. Teksty mog zawiera szablony - te bd rozwijane rekurencyjnie.
		double[] limits = { 0, 1, 2 };
		String[] fileStrings = { rb1.getString("noFiles"), rb1.getString("oneFile"), rb1.getString("multipleFiles") };
		// Potem zostanie utworzona instancja ChoiceFormat
		ChoiceFormat cf = new ChoiceFormat(limits, fileStrings);
		// Teraz z bundla mona wczyta szablon, tutaj to tekst: "There {0} on {1}"
		String pattern = rb1.getString("pattern"); 
		// Nastpnie naley okreli formatowanie, jakie ma by zastosowane do kolejnych
		// atrybutw. Robi si to za pomoc tabeli.
		// Tutaj tabela ma 3 elementy, cho w szablonie mamy tylko dwa indeksy:
		/// {0} i {1}. Trzeci indeks {2} pojawi si w wyniku formatowania dokonanego
		// przez cf (z wartoci > 2) - patrz dalej
		Format[] formats = { cf, null, NumberFormat.getInstance() };
		// formatowany bdzie szablon "There {0} on {1}", bo tak ustawiono to w mf
		mf.applyPattern(pattern);
		// Formatowanie pozycji w szablonie dokonywa bd instacje z tabeli formats,
		// przy czym, jak ju wspomniano, do sformatowania s dwie indeksowane pozycjie,
		// a instancji formatujcych jest 3.
		mf.setFormats(formats);
		// Trzecia instancja odpali si dopiero przy formatowaniu dla wartoci > 2
		// wtedy cf zwrci "are {2} files" - czyli cig znakw zawierajcy {2}.
		// Taki cig znakw bdzie formatowany dalej (rekurencyjnie), a w efekcie odpali
		// si trzeci formater: NumberFormat
		Object[] ma = { null, "Disk", null };
		for (int numFiles = 0; numFiles < 4; numFiles++) {
			ma[0] = numFiles;
			ma[2] = numFiles;
			String result = mf.format(ma);
			System.out.println(result);
		}

		/* Istnieje moliwo okrelenia szablonw w kodzie programu
		 *  https://docs.oracle.com/javase/7/docs/api/java/text/MessageFormat.html
		 *  
		 *   mf.applyPattern("There {0,choice,0#are no files|1#is one file|1<are {0,number,integer} files}.");

		 * Ponadto: 
		 * - przy zdefiniowanym szablonie z zadeklarowanym formatowaniem mona 
		 *   w niego wstrzykn dane;
		 * - dla danego cigu znakw mona wykorzysta szablon, by z tego cigu znakw 
		 *   sparsowa warto w pozycji odpowiednio do przekazanego indeksu 
		 *  
		 *  	MessageFormat mf = new
		 * 		MessageFormat("{0,number,#.##}, {0,number,#.#}"); 
		 * 		Object[] objs = {new Double(3.1415)}; 
		 * 		String result = mf.format( objs ); // result now equals "3.14, 3.1" 
		 * 		objs = null; 
		 * 		objs = mf.parse(result, new ParsePosition(0)); // objs now equals {new Double(3.1)}
		 *
		 *
		 * Z parsowaniem z uyciem szablonw wie si jeden problem:
		 * - jeli w cigu formatujcym dany indeks wystpi kilka razy, 
		 *   to sparsowana zostanie warto z miejsca ostatniego wystpienie tego indeksu
		 *   
		 *    MessageFormat mf = new MessageFormat("{0}, {0}, {0}"); 
		 *    String forParsing = "x, y, z"; 
		 *    Object[] objs = mf.parse(forParsing, new ParsePosition(0)); // result now equals {new String("z")}
		 * 
		 * Formatowanie dat i wartoci:
		 * 
		 * W klasach pomocniczych NumberFormat
		 */
		// Wspierane lokalizacje
		// Locale[] locales = NumberFormat.getAvailableLocales();
        // for(Locale lc : locales)
        //	System.out.println(lc);
		Double amount = 345678.246;
		NumberFormat nf = NumberFormat.getNumberInstance(l);
		String out = nf.format(amount);
		System.out.println(out);
		nf = NumberFormat.getCurrencyInstance(l);
		out = nf.format(amount);
		System.out.println(out);
		amount = 0.75;
		nf = NumberFormat.getPercentInstance(l);
		out = nf.format(amount);
		System.out.println(out);
		Date today = new Date();
		DateFormat df = DateFormat.getDateInstance(DateFormat.DEFAULT,l);
		DateFormat tf = DateFormat.getTimeInstance(DateFormat.DEFAULT, l);
		DateFormat dtf = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG, l);
		System.out.println(df.format(today) + ", " +
							tf.format(today) + ", " +
							 dtf.format(today));
		/* character API:
		 * isDigit 
		 * isLetter 
		 * isLetterOrDigit 
		 * isLowerCase 
		 * isUpperCase 
		 * isSpaceChar 
		 * isDefined
         
         Metody tego API stosuje si do operacji na znakach (zamiast operowa na ich kodach)
         if (Character.isLetter(ch)) ....
         
         Niektre znaki lub ich charakter reprezentowane s przez stae:
         LOWERCASE_LETTER, UPPERCASE_LETTER, MATH_SYMBOL, CONNECTOR_PUNCTUATION, ...
         	
         if (Character.getType('_') == Character.CONNECTOR_PUNCTUATION)
             // ch is a "connector"
         
         Zbiory znakw podlegaj rnym reguom sortowania, podziau itp. w zalenoci od lokalizacji
         Aby nad tym zapanowa korzysta si z klas:
          
         Collator, RuleBasedCollator, CollationKey

        */ 

		Collator en_UScol = Collator.getInstance(new Locale("en","US"));
        Collator fr_FRcol = Collator.getInstance(new Locale("fr","FR"));
        
        List<String> ls = new ArrayList<>(Arrays.asList(new String[]{"peach", "p\u00EAche", "p\u00E9ch\u00E9", "sin"}));
        ls.sort(en_UScol);
        System.out.println(ls);
        ls.sort(fr_FRcol);
        System.out.println(ls);
        
        
        /* Do rozpoznawania przerw w tekstach stosuje si klas BreakIterator 
         * Pozwala ona na stworzenie wirtualnego kursora, ktry ustawia si bdzie,
         * odpowiednio do zastosowanych funkcji
		 * - na granicach znaku: getCharacterInstance
		 * - na granicach sw: getWordInstance
		 * - na granicach zda: getSentenceInstance
		 * - na granicach linii: getLineInstance

https://docs.oracle.com/javase/7/docs/api/java/text/BreakIterator.html

Line boundary analysis determines where a text string can be broken when line-wrapping. 
The mechanism correctly handles punctuation and hyphenated words. 
Actual line breaking needs to also consider the available line width and 
is handled by higher-level software.

Sentence boundary analysis allows selection with correct interpretation of periods 
within numbers and abbreviations, and trailing punctuation marks such as quotation 
marks and parentheses.

Word boundary analysis is used by search and replace functions, as well as within 
text editing applications that allow the user to select words with a double click. 
Word selection provides correct interpretation of punctuation marks within and 
following words. Characters that are not part of a word, such as symbols or punctuation 
marks, have word-breaks on both sides.

Character boundary analysis allows users to interact with characters as they expect to, 
for example, when moving the cursor through a text string. 
Character boundary analysis provides correct navigation through character strings, 
regardless of how the character is stored. The boundaries returned may be those 
of supplementary characters, combining character sequences, or ligature clusters. 
For example, an accented character might be stored as a base character and a diacritical 
mark. What users consider to be a character can differ between languages. 

        */
       
        
     	BreakIterator iterator = BreakIterator.getWordInstance(l);
     	String in =	"Zobaczmy, jak to dziaa.     " + 
     	        "W drugim zdaniu wstawimy cytat: \"Tutaj tekst cytatu\""; 
        StringBuffer markers = new StringBuffer(); 
        markers.setLength(in.length() + 1); 
        for (int k = 0; k < markers.length(); k++) 
         markers.setCharAt(k,' ');
       
        iterator.setText(in);
        int boundary = iterator.first();
        while (boundary != BreakIterator.DONE) { 
        	markers.setCharAt(boundary, '^');
        	boundary = iterator.next();
        }
        System.out.println(in); 
        System.out.println(markers);
                
	}

}

/**
 * Uwaga: projekt skompilowano z wykorzystaniem jdk-11.0.1 i spakowano do bundles.jar
 * Prosz zwrci uwag na struktur plikw w tym pliku jar.
 * 
 * Klas Main mona uruchomi z linii komend w nastpujcy sposb:
 * >...\jdk-11.0.1\bin\java.exe --module-path bundles.jar -m bundles/example.Main 
 * lub 
 * >...\jdk-11.0.1\bin\java.exe -p bundles.jar -m bundles/example.Main 
 * 
Wynik dziaania:

czoem
jak si masz?
do widzenia
czoem C
jak si masz C?
do widzenia C
At 23:29 on 11 marca 2020, we attack the 7 ships on planet Saturn.
There are no files on Disk.
There is one file on Disk.
There are 2 files on Disk.
There are 3 files on Disk.
345678,246
345678,25z
75%
11 mar 2020, 23:29:01, 11 marca 2020 23:29:01 CET
[peach, pch, p?che, sin]
[peach, p?che, pch, sin]
Zobaczmy, jak to dziaa.     W drugim zdaniu wstawimy cytat: "Tutaj tekst cytatu"
^       ^^^  ^^ ^^     ^^    ^^^     ^^     ^^       ^^    ^^^^    ^^    ^^     ^^
 * 
 *  
 */
